/* Copyright 2001 Neil Edelman, distributed under the terms of the
 GNU General Public License, see copying.txt */

#include <windows.h>
#include "res.h"
#include "main.h"

int PASCAL WinMain(HINSTANCE hinst, HINSTANCE hinstPrev, LPSTR lpCmdLine, int nCmdShow) {
	HANDLE hHeap;
	int nErrorLevel;

	/* get a private heap for memory */
	if(!(hHeap = HeapCreate(0, 0, 0))) {
		MessageBox(NULL, "Windows, for unknown reasons, isn't letting the program initialize its memory heap.", "Oh, This is Just Great", MB_OK | MB_ICONERROR);
		return 2;
	}
	/* send required info to the dialog */
	{
		void *lpInfo[2];

		/* fill in the info */
		lpInfo[0] = hinst;
		lpInfo[1] = lpCmdLine;
		/* open the dialog box and save its return-value */
		nErrorLevel = DialogBoxParam(hinst, MAKEINTRESOURCE(IDD_MAINDIALOG), NULL, (DLGPROC)DialogFunc, (LPARAM)lpInfo);
	}
	/* kill that heap */
	if(!HeapDestroy(hHeap)) {
		MessageBox(NULL, "It's most unusual, but Windows isn't letting the program destroy its memory heap.", "This is NOT Normal", MB_OK | MB_ICONWARNING);
		return 1;
	}
	/* end the program */
	return nErrorLevel;
}

static int FAR PASCAL DialogFunc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	static HINSTANCE hinst = NULL;
	static BOOL bReady = FALSE, bClobberExisting = FALSE;

	switch (msg) {
		/* FIXME: how do I accept drag-and-drop files? */
	case WM_INITDIALOG:
		/* FIXME: how do I set a dialog box icon? */
		/* set some limits to the input field lengths */
		SendDlgItemMessage(hwndDlg, ID_INPUTFILENAME, EM_LIMITTEXT, MAX_PATH - 1, 0);
		SendDlgItemMessage(hwndDlg, ID_INPUTPASSWORD, EM_LIMITTEXT, SHRT_MAX - 1, 0);
		SendDlgItemMessage(hwndDlg, ID_OUTPUTFILENAME, EM_LIMITTEXT, MAX_PATH - 1, 0);
		/* these two buttons start off disabled */
		EnableWindow(GetDlgItem(hwndDlg, ID_ENCRYPT), FALSE);
		EnableWindow(GetDlgItem(hwndDlg, ID_DECRYPT), FALSE);
		/* use any command-line arguments */
		{
			char *lpCmdLine;

			/* fill in the info */
			hinst = ((void **)lParam)[0];
			lpCmdLine = ((void **)lParam)[1];
			/* if there were arguments */
			if(lpCmdLine && *lpCmdLine) {
				BOOL bEncrypt = FALSE, bDecrypt = FALSE, bUseNull = FALSE, bTerminate = FALSE, bHelp = FALSE;
				char *lpcArg = lpCmdLine, *szInFile = NULL, *szOutFile = NULL, *szPword = NULL;

				/* skip initial whitespce */
				while(*lpcArg == ' ' || *lpcArg == *"\t") lpcArg++;
				/* while there is a switch */
				while(*lpcArg == '/' || *lpcArg == '-') {
					/* skip past all switch characters */
					while(*lpcArg == '/' || *lpcArg == '-') lpcArg++;
					/* if the switch actually exists and is one letter long */
					if(*lpcArg && (lpcArg[1] == ' ' || lpcArg[1] == *"\t" || !lpcArg[1])) {
						if(*lpcArg == 'e') bEncrypt = TRUE;
						else if(*lpcArg == 'd') bDecrypt = TRUE;
						else if(*lpcArg == 'n') bUseNull = TRUE;
						else if(*lpcArg == 't') bTerminate = TRUE;
						else if(*lpcArg == 'o') bClobberExisting = TRUE;
						else if(*lpcArg == '?' || *lpcArg == 'h') bHelp = TRUE;
					}
					/* skip past the switch */
					while(*lpcArg && *lpcArg != ' ' && *lpcArg != *"\t") lpcArg++;
					/* skip to the next arg */
					while(*lpcArg == ' ' || *lpcArg == *"\t") lpcArg++;
				}
				/* get the first argument (input file path) */
				szInFile = lpcArg;
				/* skip to the end of the arg */
				while(*lpcArg && *lpcArg != ' ' && *lpcArg != *"\t") lpcArg++;
				/* if there's more */
				if(*lpcArg) {
					/* terminate the string here */
					*lpcArg = 0;
					/* move on */
					lpcArg++;
					/* skip whitespce */
					while(*lpcArg == ' ' || *lpcArg == *"\t") lpcArg++;
					/* the next argument is the output file path */
					szOutFile = lpcArg;
					/* skip to the end of the arg */
					while(*lpcArg && *lpcArg != ' ' && *lpcArg != *"\t") lpcArg++;
					/* test for further arguments */
					if(*lpcArg) {
						/* terminate the string here */
						*lpcArg = 0;
						/* move on */
						lpcArg++;
						/* the next arg is the password */
						szPword = lpcArg;
					}
				}
				/* initialize the fields with the command string */
				{
					signed short sFieldShift = 0;

					if(szInFile && *szInFile) {
						/* set the input field */
						SetDlgItemText(hwndDlg, ID_INPUTFILENAME, szInFile);
						/* move to the password field */
						sFieldShift = 1;
					}
					if(szOutFile && *szOutFile) {
						/* set the output field */
						SetDlgItemText(hwndDlg, ID_OUTPUTFILENAME, szOutFile);
						/* move to the password field */
						sFieldShift = 1;
					}
					if(szPword && *szPword) {
						/* set the password field */
						SetDlgItemText(hwndDlg, ID_INPUTPASSWORD, szPword);
						/* move to the use null checkbox */
						sFieldShift = 2;
					}
					/* move the specified number of fields ahead */
					while(sFieldShift > 0) {
						/* set the focus to the next field */
						PostMessage(hwndDlg, WM_NEXTDLGCTL, 0, FALSE); /* can't figure out how to get it to go to a specific control . . . documentation is FUBAR on this */
						/* decrement counter */
						sFieldShift--;
					}
				}
				if(bHelp) SendMessage(hwndDlg, WM_COMMAND, ID_ABOUT, 0); /* the actual message is not checked in this program */
				/* check for the null flag */
				if(bUseNull) CheckDlgButton(hwndDlg, ID_USENULL, TRUE);
				/* check decrypt flag */
				if(bDecrypt && !bEncrypt) SendMessage(hwndDlg, WM_COMMAND, ID_DECRYPT, 0); /* the actual message is not checked in this program */
				/* check encrypt flag */
				else if(bEncrypt && !bDecrypt) SendMessage(hwndDlg, WM_COMMAND, ID_ENCRYPT, 0); /* the actual message is not checked in this program */
				/* check terminate flag */
				if(bTerminate) EndDialog(hwndDlg, 0);
			}
		}
		/* we did something */
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam) /* WIN32 CTLID */) {
		case ID_INPUTFILENAME:
			if(HIWORD(wParam) /* WIN32 CTLMSG */ == EN_UPDATE) {
				int nInFileLen, end = 0, i;
				char *szInFile;

				/* find the text length */
				nInFileLen = SendDlgItemMessage(hwndDlg, ID_INPUTFILENAME, WM_GETTEXTLENGTH, 0, 0);
				if(nInFileLen) {
					/* try getting the space for the text data, plus the maximum 4 extra for an added extension */
					if((szInFile = Alloc(hwndDlg, nInFileLen + 5))) {
						/* get the text data */
						GetDlgItemText(hwndDlg, ID_INPUTFILENAME, szInFile, nInFileLen + 1);
						/* erase the extension and add a default extension */
						for(i = 0; i < nInFileLen && szInFile[i]; i++) {
							if(szInFile[i] == '.') end = i;
						}
						if(end == 0) end = i;
						if(szInFile[end] == '.' && szInFile[end + 1] == '_') {
							szInFile[++end] = 'f';
							szInFile[++end] = 'o';
							szInFile[++end] = 'o';
							szInFile[++end] = 0;
						}
						else {
							szInFile[end++] = '.';
							szInFile[end++] = '_';
							szInFile[end] = 0;
						}
						/* set this as the output file name */
						SetDlgItemText(hwndDlg, ID_OUTPUTFILENAME, szInFile);
						/* free the text */
						Free(hwndDlg, &szInFile);
					}
				}
				else {
					/* empty output file name */
					SetDlgItemText(hwndDlg, ID_OUTPUTFILENAME, "");
				}
			}
			/* continue with the rest of the checks */
		case ID_INPUTPASSWORD:
		case ID_OUTPUTFILENAME:
			if(HIWORD(wParam) /* WIN32 CTLMSG */ == EN_UPDATE) {
				/* if all of the fields are full */
				if(	SendDlgItemMessage(hwndDlg, ID_INPUTFILENAME, WM_GETTEXTLENGTH, 0, 0) &&
					SendDlgItemMessage(hwndDlg, ID_INPUTPASSWORD, WM_GETTEXTLENGTH, 0, 0) &&
					SendDlgItemMessage(hwndDlg, ID_OUTPUTFILENAME, WM_GETTEXTLENGTH, 0, 0)) {
					/* if this is newly ready */
					if(!bReady) {
						/* enable the buttons */
						EnableWindow(GetDlgItem(hwndDlg, ID_ENCRYPT), TRUE);
						EnableWindow(GetDlgItem(hwndDlg, ID_DECRYPT), TRUE);
						/* we're ready */
						SetDlgItemText(hwndDlg, IDT_STATUS, "ready");
						/* now ready */
						bReady = TRUE;
					}
				}
				else {
					/* not ready anymore */
					if(bReady) {
						/* disable buttons */
						EnableWindow(GetDlgItem(hwndDlg, ID_ENCRYPT), FALSE);
						EnableWindow(GetDlgItem(hwndDlg, ID_DECRYPT), FALSE);
						/* not ready */
						SetDlgItemText(hwndDlg, IDT_STATUS, "need input");
						/* not ready */
						bReady = FALSE;
					}
				}
			}
			/* we did something */
			return TRUE;
		case ID_ENCRYPT:
		case ID_DECRYPT:
			{
				HANDLE hFSource = INVALID_HANDLE_VALUE, hFTarget = INVALID_HANDLE_VALUE; /* in case of error */
				int nInFileLen, nPwordLen, nOutFileLen, nExtraNull;
				char *szInFile = NULL, *szPword = NULL, *szOutFile = NULL; /* NULL so they can be checked in case of memory error */

				/* get the lengths of each field ( +1 b/c the \0 on the end) */
				nInFileLen = SendDlgItemMessage(hwndDlg, ID_INPUTFILENAME, WM_GETTEXTLENGTH, 0, 0);
				nPwordLen = SendDlgItemMessage(hwndDlg, ID_INPUTPASSWORD, WM_GETTEXTLENGTH, 0, 0);
				nOutFileLen = SendDlgItemMessage(hwndDlg, ID_OUTPUTFILENAME, WM_GETTEXTLENGTH, 0, 0);
				/* allocate space for the text */
				if(!(szInFile = Alloc(hwndDlg, nInFileLen + 1)) || !(szPword = Alloc(hwndDlg, nPwordLen + 1)) || !(szOutFile = Alloc(hwndDlg, nOutFileLen + 1))) {
					Free(hwndDlg, &szInFile);
					Free(hwndDlg, &szPword);
					Free(hwndDlg, &szOutFile);
				}
				/* get the text from the dialog items */
				GetDlgItemText(hwndDlg, ID_INPUTFILENAME, szInFile, nInFileLen + 1);
				GetDlgItemText(hwndDlg, ID_INPUTPASSWORD, szPword, nPwordLen + 1);
				GetDlgItemText(hwndDlg, ID_OUTPUTFILENAME, szOutFile, nOutFileLen + 1);
				/* check that the length of the password is within limits */
				if(nPwordLen < 1) {
					/* free the name data */
					Free(hwndDlg, &szInFile);
					Free(hwndDlg, &szPword);
					Free(hwndDlg, &szOutFile);
					/*report the error */
					MessageBox(hwndDlg, "Password character sequence must exist.", "Can't Work With This", MB_OK | MB_ICONASTERISK);
					SetDlgItemText(hwndDlg, IDT_STATUS, "missing data");
					return TRUE;
				}
				if(nPwordLen > BIG_PASSWORD_LENGTH && MessageBox(hwndDlg, "That password looks very long; are you sure that that's what you want?", "This isn't a Dissertation", MB_YESNO | MB_ICONQUESTION) == IDNO) {
					/* free the name data */
					Free(hwndDlg, &szInFile);
					Free(hwndDlg, &szPword);
					Free(hwndDlg, &szOutFile);
					/*report the "error" */
					SetDlgItemText(hwndDlg, IDT_STATUS, "operation aborted");
					return TRUE;
				}
				/* attempt to load the specified stuff */
				if(!LoadFiles(hwndDlg, szInFile, &hFSource, szOutFile, &hFTarget, bClobberExisting)) {
					/* close any open files */
					CloseFiles(hwndDlg, &hFSource, &hFTarget);
					/* free the name data */
					Free(hwndDlg, &szInFile);
					Free(hwndDlg, &szPword);
					Free(hwndDlg, &szOutFile);
					/* the message was handled */
					return TRUE;
				}
				/* see if the null is to be included */
				if(IsDlgButtonChecked(hwndDlg, ID_USENULL)) nExtraNull = 1;
				else nExtraNull = 0;
				/* disable input */
				EnableWindow(GetDlgItem(hwndDlg, ID_INPUTFILENAME), FALSE);
				EnableWindow(GetDlgItem(hwndDlg, ID_INPUTPASSWORD), FALSE);
				EnableWindow(GetDlgItem(hwndDlg, ID_OUTPUTFILENAME), FALSE);
				EnableWindow(GetDlgItem(hwndDlg, ID_ENCRYPT), FALSE);
				EnableWindow(GetDlgItem(hwndDlg, ID_DECRYPT), FALSE);
				EnableWindow(GetDlgItem(hwndDlg, ID_ABOUT), FALSE);
				EnableWindow(GetDlgItem(hwndDlg, IDCANCEL), FALSE);
				/* find out if this is an attempt at encrypting or decrypting */
				if(LOWORD(wParam) == ID_ENCRYPT) {
					/* send this info to the encryption function */
					Encrypt(hwndDlg, hFSource, szPword, nPwordLen + nExtraNull, hFTarget);
				}
				else {
					/* send this info to the decryption function */
					Decrypt(hwndDlg, hFSource, szPword, nPwordLen + nExtraNull, hFTarget);
				}
				/* free the name data */
				Free(hwndDlg, &szInFile);
				Free(hwndDlg, &szPword);
				Free(hwndDlg, &szOutFile);
				/* close the opened files */
				CloseFiles(hwndDlg, &hFSource, &hFTarget);
				/* re-enable input as appropriate */
				EnableWindow(GetDlgItem(hwndDlg, ID_INPUTFILENAME), TRUE);
				EnableWindow(GetDlgItem(hwndDlg, ID_INPUTPASSWORD), TRUE);
				EnableWindow(GetDlgItem(hwndDlg, ID_OUTPUTFILENAME), TRUE);
				if(bReady) {
					EnableWindow(GetDlgItem(hwndDlg, ID_ENCRYPT), TRUE);
					EnableWindow(GetDlgItem(hwndDlg, ID_DECRYPT), TRUE);
				}
				EnableWindow(GetDlgItem(hwndDlg, ID_ABOUT), TRUE);
				EnableWindow(GetDlgItem(hwndDlg, IDCANCEL), TRUE);
				/* the input focus usually just disappears from everything here (for no reason?), so get it back */
				SendMessage(hwndDlg, WM_SETFOCUS, 0, 0);
			}
			/* we did something */
			return TRUE;
		case ID_ABOUT:
			/* show the dialog */
			DialogBox(hinst, MAKEINTRESOURCE(IDD_ABOUTDIALOG), hwndDlg, AboutDialogFunc);
			/* we did something */
			return TRUE;
		case IDOK:
			{
				int nInFileLen, end = 0, i;
				char *szInFile;

				/* find the text length of the input file name */
				nInFileLen = SendDlgItemMessage(hwndDlg, ID_INPUTFILENAME, WM_GETTEXTLENGTH, 0, 0);
				/* if there is text in this field, as well as the other required fields */
				if(nInFileLen && SendDlgItemMessage(hwndDlg, ID_INPUTPASSWORD, WM_GETTEXTLENGTH, 0, 0) && SendDlgItemMessage(hwndDlg, ID_OUTPUTFILENAME, WM_GETTEXTLENGTH, 0, 0)) {
					/* try getting the space for the text data */
					if((szInFile = Alloc(hwndDlg, nInFileLen + 1))) {
						/* get the text data */
						GetDlgItemText(hwndDlg, ID_INPUTFILENAME, szInFile, nInFileLen + 1);
						/* find the end of the path */
						for(i = 0; i < nInFileLen && szInFile[i]; i++) {
							if(szInFile[i] == '.') end = i;
						}
						if(end == 0) end = i;
						/* if the file has the default encrypted extension */
						if(szInFile[end] == '.' && szInFile[end + 1] == '_') {
							/* see if the user wants to decrypt the file */
							if(MessageBox(hwndDlg, "DECRYPT the input file; is this correct?", "Uncertain", MB_YESNO | MB_ICONQUESTION) == IDYES) {
								/* try it */
								SendMessage(hwndDlg, WM_COMMAND, ID_DECRYPT, 0); /* the actual message is not checked in this program */
							}
						}
						else {
							/* see if the user wants to encrypt the file */
							if(MessageBox(hwndDlg, "ENCRYPT the input file; is this correct?", "Uncertain", MB_YESNO | MB_ICONQUESTION) == IDYES) {
								/* try it */
								SendMessage(hwndDlg, WM_COMMAND, ID_ENCRYPT, 0); /* the actual message is not checked in this program */
							}
						}
						/* free the text */
						Free(hwndDlg, &szInFile);
					}
				}
			}
			/* the message was handled */
			return TRUE;
		case IDCANCEL:
			/* quit */
			EndDialog(hwndDlg, 0);
			/* we did something */
			return TRUE;
		}
		/* just in case we get some wierded message */
		break;
	case WM_CLOSE:
		/* quit */
		EndDialog(hwndDlg, 0);
		/* we did something */
		return TRUE;
	}
	/* we did nothing */
	return FALSE;
}

static int FAR PASCAL AboutDialogFunc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch (msg) {
	case WM_COMMAND:
		switch (LOWORD(wParam) /* WIN32 CTLID */) {
		case ID_HELP:
				MessageBox(hwndDlg, "This is an encryption programme.\n\n\
The \"append terminating null\" option adds a null character to the end of \
your password; the previous version of this program always did this. Because \
the null is part of the password, this toggle must be set to the same state \
when decrypting as it was when encrypting.", "Help", MB_OK);
			return TRUE;
		case ID_LEGAL:
				MessageBox(hwndDlg, "Copyright (C) 2001 Neil Edelman, distributed \
under the terms of the GNU General Public License.\n\n\
Version 2.0.\n\n\
Obfuscator is free software: you can redistribute it and/or modify \
it under the terms of the GNU General Public License as published by \
the Free Software Foundation, either version 3 of the License, or \
(at your option) any later version.\n\n\
This program is distributed in the hope that it will be useful, \
but WITHOUT ANY WARRANTY; without even the implied warranty of \
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the \
GNU General Public License for more details, \
<http://www.gnu.org/licenses/>.", "Copyright", MB_OK);
			return TRUE;
		case ID_INFO:
				MessageBox(hwndDlg, "The encryption process uses an stream \
cipher algorithm. The length of the password is linearly related to how many \
bits of encryption. The bits of the key generate a pseudorandom keysteam. It \
is possible that this is insecure.", "Encryption", MB_OK);
			return TRUE;
		case ID_SYNTAX:
			MessageBox(hwndDlg, "Usage:\nObfuscator-2_1.exe [<options>] [<input> [<output> [<password>]]]\n\n\
Where:\n<options> - any combination of:\n\t-e (start encryption immediately);\n\t-d (start decryption immediately, negates -e);\n\t-n (add null character to password);\n\t-t \
(terminate immediately);\n\t-o (overwrite existing files without warning); and\n\t-? or -h (show info dialog).\n<output> - the output file path; and\n<password> - the encryption \
character sequence.\n\n\
The password starts exactly one space after <output> and continues to the end, to facilitate leading or trailing spaces. Real paths must be used (i.e. no spaces).\n\n\
e.g.:\nObfuscator-2_1 -e -t -o c:\\mydocu~1\\myfile.txt c:\\mydocu~1\\myfile._ bar\n\n\
Try [HKEY_CLASSES_ROOT\\Unknown\\shell\\obfuscate\\command]@=\"Obfuscator-2_1.exe %1\". \
Dragging a file onto the program icon will also work.", "Syntax", MB_OK);
			return TRUE;
		case ID_AUTHOR:
			MessageBox(hwndDlg, "This program was written by Neil Edelman, <http://neil.chaosnet.org/>.", "Author", MB_OK);
			return TRUE;
		/* FIXME: make the last button selected */
		case IDOK:
		case IDCANCEL:
			EndDialog(hwndDlg, 0);
			return TRUE;
		}
	}
	return FALSE;
}

/* FIXME: put a progress bar in a separate windows to eliminate the problem of unresponsiveness during work */

void Encrypt(HWND hwndDlg, HANDLE hFSource, char *szPword, int nPwordLen, HANDLE hFTarget) {
	DWORD nNumberOfBytes = 1;
	int nPwordPos = 0, nKiloBytesRemaining, i;
	char szStatus[256];
	byte prevDataParity = 0, key, data, addr, garb, valueTable[256], addressTable[256];

	/* check the parameters sent to the function */
	if(!hFSource || !szPword || !nPwordLen || !hFTarget) {
		MessageBox(hwndDlg, "Program Error: some input data didn't make it to the process.", "Oh Dear", MB_OK | MB_ICONERROR);
		SetDlgItemText(hwndDlg, IDT_STATUS, "program error");
		return;
	}
	/* make a table of translation values */
	{
		int shiftVal = 0;

		/* determine how far the initial values in the table should be shifted */
		for(i = 0; i < nPwordLen; i++) {
			shiftVal ^= szPword[i] << (i & 7);
		}
		shiftVal -= nPwordLen; /* if the password is "xxxxx" this makes it more powerful than it otherwise would be (decypting with "xxx" might otherwise yield a less-strongly encrypted file_ */
		/* write out the table which determines the values to which each data value is mapped and a table for their addresses */
		for(i = 0; i < 256; i++) {
			valueTable[i] = (byte)(i + shiftVal);
			addressTable[i] = (byte)(i - shiftVal); /* shifting values one way shifts the address of the values the other way */
		}
	}
	/* find out how big the file is */
	nKiloBytesRemaining = (GetFileSize(hFSource, NULL) >> 10) + 1; /* round up to make the program seem faster :P */
	/* go through and mess with the file */
	while(nNumberOfBytes) {
		wsprintf(szStatus, "kilobytes remaining: %d", nKiloBytesRemaining);
		SetDlgItemText(hwndDlg, IDT_STATUS, szStatus);
		nKiloBytesRemaining--;
		/* go through the next kilobyte */
		for(i = 0; i < 1024; i++) {
			/* get a byte of input */
			if(!ReadFile(hFSource, &data, sizeof(byte), &nNumberOfBytes, NULL)) {
				MessageBox(hwndDlg, "The input file spewed forth vile garbage! Output is unreliable.", "The End is Here", MB_OK | MB_ICONERROR);
				SetDlgItemText(hwndDlg, IDT_STATUS, "file i/o error");
				return;
			}
			/* make sure that something was read (i.e. not @ EOF) */
			if(!nNumberOfBytes) break;
			/* make sure that the correct amount was read */
			if(nNumberOfBytes != sizeof(byte)) {
				MessageBox(NULL, "A peculiar file I/O error has resulted in the wrong amount of data being read. Output is unreliable.", "Ohno", MB_OK | MB_ICONERROR);
				SetDlgItemText(hwndDlg, IDT_STATUS, "file i/o error");
				return;
			}
			/* get the address in the valuetable of the data to write */
			addr = addressTable[data];
			/* create an encryption seed based on the possword and file position */
			{
				int nGlyphsUsed;
				char *lpChar;

				key = 0;
				lpChar = szPword + nPwordPos; /* address of the current character in the password */
				for(nGlyphsUsed = 0; nGlyphsUsed < nPwordLen; nGlyphsUsed++) {
					if(!*lpChar) lpChar = szPword; /* if at the end of the sequence, loop back to the start */
					key ^= (*lpChar) << (nGlyphsUsed & 7); /* use the shifting because normal passwords only contain alphanumeric values, so it is now possible to have any byte value */
				}
				key += ((byte)i ^ (byte)(i >> (szPword[nPwordLen - 1 - nPwordPos] & 7))); /* so that the key changes for the same glyph in the password across the file */
			}
			/* encrypt the data to find where to look in the table */
			garb = addr ^ key;
			/* use two parts of the password, and modify the operation based on previous data */
			if(prevDataParity) garb += szPword[nPwordLen - 1 - nPwordPos];
			else garb -= szPword[nPwordLen - 1 - nPwordPos];
			/* set the parity for next time */
			prevDataParity = data & 1;
			/* now swap this result with the value of the original data in the translation table (can't use XOR b/c they're references, so must do it the slow way) */
			{
				byte dataValue = valueTable[data];

				valueTable[data] = valueTable[garb];
				valueTable[garb] = dataValue;
			}
			/* and swap the addresses which point to these values */
			{
				byte dataValueAddress = addressTable[valueTable[data]];

				addressTable[valueTable[data]] = addressTable[valueTable[garb]];
				addressTable[valueTable[garb]] = dataValueAddress;
			}
			/* dynamically modify the password sequence */
			szPword[nPwordPos] += valueTable[addr];
			/* move to the next character in the sequence */
			nPwordPos = (nPwordPos + 1) % nPwordLen;
			/* write to the output file */
			if(!WriteFile(hFTarget, &garb, sizeof(byte), &nNumberOfBytes, NULL)) {
				MessageBox(hwndDlg, "A write error caused the output file to be twisted into uselessly unintelligable mush!", "Oh, Great", MB_OK | MB_ICONERROR);
				SetDlgItemText(hwndDlg, IDT_STATUS, "file i/o error");
				return;
			}
			if(nNumberOfBytes != sizeof(byte)) {
				MessageBox(hwndDlg, "The output file has been bollixed by some eldritch file I/O singulatity.", "This is Peculiar", MB_OK | MB_ICONERROR);
				SetDlgItemText(hwndDlg, IDT_STATUS, "file i/o error");
				return;
			}
		}
	}
	/* set the focus to "done" */
	PostMessage(hwndDlg, WM_NEXTDLGCTL, 1, FALSE); /* 1 sets it there . . . nothing else seems to work (specifically, IDCANCEL does squat) - I'll never understand Windoze */
	/* admit what has been done */
	SetDlgItemText(hwndDlg, IDT_STATUS, "encryption appears successful");
}

void Decrypt(HWND hwndDlg, HANDLE hFSource, char *szPword, int nPwordLen, HANDLE hFTarget) {
	DWORD nNumberOfBytes = 1;
	int nPwordPos = 0, nKiloBytesRemaining, i;
	char szStatus[256];
	byte prevDataParity = 0, key, data, addr, garb, valueTable[256];

	/* check the parameters sent to the function */
	if(!hFSource || !szPword || !nPwordLen || !hFTarget) {
		MessageBox(hwndDlg, "Program Error: some input data didn't make it to the process.", "Oh Dear", MB_OK | MB_ICONERROR);
		SetDlgItemText(hwndDlg, IDT_STATUS, "program error");
		return;
	}
	/* make a table of translation values */
	{
		int shiftVal = 0;

		/* determine how far the initial values in the table should be shifted */
		for(i = 0; i < nPwordLen; i++) {
			shiftVal ^= szPword[i] << (i & 7);
		}
		shiftVal -= nPwordLen; /* if the password is "xxxxx" this makes it more powerful than it otherwise would be (decypting with "xxx" might otherwise yield a less-strongly encrypted file_ */
		/* write out the table which determines the values to which each data value is mapped */
		for(i = 0; i < 256; i++) {
			valueTable[i] = (byte)(i + shiftVal);
		}
	}
	/* find out how big the file is */
	nKiloBytesRemaining = GetFileSize(hFSource, NULL) >> 10;
	/* go through and mess with the file */
	while(nNumberOfBytes) {
		wsprintf(szStatus, "kilobytes remaining: %d", nKiloBytesRemaining);
		SetDlgItemText(hwndDlg, IDT_STATUS, szStatus);
		nKiloBytesRemaining--;
		/* go through the next kilobyte */
		for(i = 0; i < 1024; i++) {
			/* get a byte of input */
			if(!ReadFile(hFSource, &garb, sizeof(byte), &nNumberOfBytes, NULL)) {
				MessageBox(hwndDlg, "The input file spewed forth vile garbage! Output is unreliable.", "The End is Here", MB_OK | MB_ICONERROR);
				SetDlgItemText(hwndDlg, IDT_STATUS, "file i/o error");
				return;
			}
			/* make sure that something was read (i.e. not @ EOF) */
			if(!nNumberOfBytes) break;
			/* make sure that the correct amount was read */
			if(nNumberOfBytes != sizeof(byte)) {
				MessageBox(NULL, "A peculiar file I/O error has resulted in the wrong amount of data being read. Output is unreliable.", "It's FUBAR", MB_OK | MB_ICONERROR);
				SetDlgItemText(hwndDlg, IDT_STATUS, "file i/o error");
				return;
			}
			/* create an encryption seed based on the possword and file position */
			{
				int nGlyphsUsed;
				char *lpChar;

				key = 0;
				lpChar = szPword + nPwordPos; /* address of the current character in the password */
				for(nGlyphsUsed = 0; nGlyphsUsed < nPwordLen; nGlyphsUsed++) {
					if(!*lpChar) lpChar = szPword; /* if at the end of the sequence, loop back to the start */
					key ^= (*lpChar) << (nGlyphsUsed & 7); /* use the shifting because normal passwords only contain alphanumeric values, so it is now possible to have any byte value */
				}
				key += ((byte)i ^ (byte)(i >> (szPword[nPwordLen - 1 - nPwordPos] & 7))); /* so that the key changes for the same glyph in the password across the file */
			}
			/* reverse this part of the obfuscaption */
			if(prevDataParity) addr = garb - szPword[nPwordLen - 1 - nPwordPos];
			else addr = garb + szPword[nPwordLen - 1 - nPwordPos];
			/* decrypt the garbage in addr to find where to look in the table */
			addr ^= key;
			/* get the output data from this point in the table */
			data = valueTable[addr];
			/* set the parity for next time */
			prevDataParity = data & 1;
			/* now swap this result with the value of the original data in the translation table */
			{
				byte dataValue = valueTable[data];

				valueTable[data] = valueTable[garb];
				valueTable[garb] = dataValue;
			}
			/* dynamically modify the password sequence */
			szPword[nPwordPos] += valueTable[addr];
			/* move to the next character in the sequence */
			nPwordPos = (nPwordPos + 1) % nPwordLen;
			/* write to the output file */
			if(!WriteFile(hFTarget, &data, sizeof(byte), &nNumberOfBytes, NULL)) {
				MessageBox(hwndDlg, "A write error caused the output file to be twisted into uselessly unintelligable mush!", "Oh, Great", MB_OK | MB_ICONERROR);
				SetDlgItemText(hwndDlg, IDT_STATUS, "file i/o error");
				return;
			}
			if(nNumberOfBytes != sizeof(byte)) {
				MessageBox(hwndDlg, "The output file has been bollixed by some eldritch file I/O singulatity.", "This is Peculiar", MB_OK | MB_ICONERROR);
				SetDlgItemText(hwndDlg, IDT_STATUS, "file i/o error");
				return;
			}
		}
	}
	/* set the focus to "done" */
	PostMessage(hwndDlg, WM_NEXTDLGCTL, 1, FALSE); /* 1 sets it there . . . nothing else seems to work (specifically, IDCANCEL does squat) - I'll never understand Windoze */
	/* admit what has been done */
	SetDlgItemText(hwndDlg, IDT_STATUS, "decryption appears successful");
}

int LoadFiles(HWND hwndDlg, char *szInFile, HANDLE *hFSource, char *szOutFile, HANDLE *hFTarget, BOOL bClobberExisting) {
	/* check the filenames sent to the function */
	if(!szInFile || !szOutFile) {
		MessageBox(hwndDlg, "Program error: the file name paths coudn't be found.", "It's All Over", MB_OK | MB_ICONWARNING);
		SetDlgItemText(hwndDlg, IDT_STATUS, "program error");
		return FALSE;
	}
	if(lstrlen(szInFile) < 1) {
		MessageBox(hwndDlg, "Input file path has not beem specified.", "This Won't Do", MB_OK | MB_ICONASTERISK);
		SetDlgItemText(hwndDlg, IDT_STATUS, "missing data");
		return FALSE;
	}
	if(lstrlen(szInFile) > MAX_PATH) {
		MessageBox(hwndDlg, "Input file path exceeds Windows' length capacity.", "Message of Unavoidable Death", MB_OK | MB_ICONERROR);
		SetDlgItemText(hwndDlg, IDT_STATUS, "data overload");
		return FALSE;
	}
	if(lstrlen(szOutFile) < 1) {
		MessageBox(hwndDlg, "There must be more to the output file path.", "Feed Me", MB_OK | MB_ICONASTERISK);
		SetDlgItemText(hwndDlg, IDT_STATUS, "missing data");
		return FALSE;
	}
	if(lstrlen(szOutFile) > MAX_PATH) {
		MessageBox(hwndDlg, "Output file path exceeds Windows' length capacity.", "Message of Imminent Doom", MB_OK | MB_ICONERROR);
		SetDlgItemText(hwndDlg, IDT_STATUS, "data overload");
		return FALSE;
	}
	/* open the input file */
	if((*hFSource = CreateFile(szInFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE) {
		MessageBox(hwndDlg, "The input file refused to open. It may be currently unreadable, or it may not even exist.", "Bad", MB_OK | MB_ICONWARNING);
		SetDlgItemText(hwndDlg, IDT_STATUS, "input file open failed");
		return FALSE;
	}
	/* open using CREATE_ALWAYS if specified */
	if(bClobberExisting) {
		if((*hFTarget = CreateFile(szOutFile, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE) {
			MessageBox(hwndDlg, "Output file won't open. It may exist and be in use or read-only, or perhaps it is on a read-only drive, or the path could be invalid, or maybe something else is wrong.", "Holy Cow", MB_OK | MB_ICONWARNING);
			SetDlgItemText(hwndDlg, IDT_STATUS, "output file open failed");
			return FALSE;
		}
	}
	/* otherwise open the output file with a overwrite warning */
	else if((*hFTarget = CreateFile(szOutFile, GENERIC_WRITE, 0, NULL, CREATE_NEW, FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE) {
		/* if it didn't work, try using CREATE_ALWAYS in case the file must be overwritten */
		if(MessageBox(hwndDlg, "Output file is not available. Do you want to attempt to overwrite any existing files with that name?", "UhOh", MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2) == IDYES) {
			if((*hFTarget = CreateFile(szOutFile, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE) {
				MessageBox(hwndDlg, "The output file still won't open", "Holy Cow", MB_OK | MB_ICONWARNING);
				SetDlgItemText(hwndDlg, IDT_STATUS, "output file open failed");
				return FALSE;
			}
		}
		/* the user didn't attempt to overwrite the file */
		else {
			SetDlgItemText(hwndDlg, IDT_STATUS, "output file open aborted");
			return FALSE;
		}
	}
	/* return sucess to the caller */
	return TRUE;
}

void CloseFiles(HWND hwndDlg, HANDLE *hFSource, HANDLE *hFTarget) {
	/* close the opened files */
	if(*hFTarget != INVALID_HANDLE_VALUE) {
		if(!CloseHandle(*hFTarget)) {
			MessageBox(hwndDlg, "The output file won't close.", "Interesting . . . ", MB_OK | MB_ICONEXCLAMATION);
			return;
		}
		*hFTarget = NULL;
	}
	if(*hFSource != INVALID_HANDLE_VALUE) {
		if(!CloseHandle(*hFSource)) {
			MessageBox(hwndDlg, "The input file won't close.", "Fascinating . . . ", MB_OK | MB_ICONEXCLAMATION);
			return;
		}
		*hFSource = NULL;
	}
}

void *Alloc(HWND hwndDlg, int size) {
	HANDLE hHeap;
	void *ptr;

	/* allocate some space */
	if(!(hHeap = GetProcessHeap()) || !(ptr = HeapAlloc(hHeap, 0, size))) {
		MessageBox(hwndDlg, "Some kind of memory error has appeared.", "Vague Warning", MB_OK | MB_ICONERROR);
		return NULL;
	}
	/* send back the allocated space */
	return ptr;
}

void Free(HWND hwndDlg, void **ptr) {
	HANDLE hHeap;

	/* we need a pointer */
	if(!*ptr) return;
	/* free the data */
	if(!(hHeap = GetProcessHeap()) || !HeapFree(hHeap, 0, *ptr)) {
		MessageBox(hwndDlg, "An attempt at freeing memory failed.", "Not a Good Sign", MB_OK | MB_ICONERROR);
	}
	/* set the pointer to NULL */
	*ptr = NULL;
}
